from __future__ import annotations

from contextlib import asynccontextmanager
from json import dumps, loads
from os.path import dirname, join
from typing import TYPE_CHECKING, Any, AsyncGenerator, cast

from aiofiles import open as aiopen
from aiofiles.os import listdir, makedirs
from aiofiles.ospath import exists as aio_exists

if TYPE_CHECKING:
    from rxon.models import FileMetadata

    from .s3 import S3Manager


class TaskFiles:
    """
    A helper class for managing task-specific files.
    Provides asynchronous lazy directory creation and high-level file operations
    within an isolated workspace for each task.
    """

    def __init__(self, task_dir: str, job_id: str, task_id: str, s3_manager: S3Manager | None = None):
        """
        Initializes TaskFiles with specific directory and identifiers.
        """
        self._task_dir = task_dir
        self._job_id = job_id
        self._task_id = task_id
        self._s3_manager = s3_manager

    async def get_root(self) -> str:
        """
        Asynchronously returns the root directory for the task.
        Creates the directory on disk if it doesn't exist.
        """
        await makedirs(self._task_dir, exist_ok=True)
        return self._task_dir

    async def path_to(self, filename: str) -> str:
        """
        Asynchronously returns an absolute path for a file within the task directory.
        Guarantees that the task root directory exists.
        """
        root = await self.get_root()
        return join(root, filename)

    def get_root_sync(self) -> str:
        """
        Synchronously returns the root directory for the task.
        Creates the directory on disk if it doesn't exist.
        """
        from os import makedirs as std_makedirs

        std_makedirs(self._task_dir, exist_ok=True)
        return self._task_dir

    def path_to_sync(self, filename: str) -> str:
        """
        Synchronously returns an absolute path for a file within the task directory.
        Guarantees that the task root directory exists.
        """
        root = self.get_root_sync()
        return join(root, filename)

    @asynccontextmanager
    async def open(self, filename: str, mode: str = "r") -> AsyncGenerator:
        """
        An asynchronous context manager to open a file within the task directory.
        Automatically creates the task root and any necessary subdirectories.

        Args:
            filename: Name or relative path of the file.
            mode: File opening mode (e.g., 'r', 'w', 'a', 'rb', 'wb').
        """
        path = await self.path_to(filename)
        # Ensure directory for the file itself exists if filename contains subdirectories
        file_dir = dirname(path)
        if file_dir != self._task_dir:
            await makedirs(file_dir, exist_ok=True)

        async with cast(Any, aiopen)(path, mode) as f:
            yield f

    async def read(self, filename: str, mode: str = "r") -> str:
        """
        Asynchronously reads the entire content of a file.

        Args:
            filename: Name of the file to read.
            mode: Mode to open the file in (defaults to 'r').
        """
        async with self.open(filename, mode) as f:
            content = await f.read()
            return cast(str, content)

    async def write(self, filename: str, data: str | bytes, mode: str = "w") -> None:
        """
        Asynchronously writes data to a file. Creates or overwrites the file by default.

        Args:
            filename: Name of the file to write.
            data: Content to write (string or bytes).
            mode: Mode to open the file in (defaults to 'w').
        """
        async with self.open(filename, mode) as f:
            await f.write(data)

    async def write_json(self, filename: str, data: Any) -> FileMetadata | None:
        """Writes data as JSON and optionally uploads to S3 if manager is available."""
        content = dumps(data, indent=2)
        await self.write(filename, content)
        if self._s3_manager:
            return await self.upload_file(filename)
        return None

    async def read_json(self, filename: str) -> Any:
        """Reads a file and parses it as JSON."""
        content = await self.read(filename)
        return loads(content)

    async def upload_file(self, filename: str) -> FileMetadata:
        """Uploads a specific file to S3 and returns its metadata."""
        if not self._s3_manager:
            raise RuntimeError("S3Manager not configured for this TaskFiles instance.")
        path = await self.path_to(filename)
        return await self._s3_manager._upload_to_s3(path, s3_prefix=self._job_id)

    async def upload_dir(self, dirname: str = "") -> FileMetadata:
        """Uploads the entire task directory or a subdirectory to S3."""
        if not self._s3_manager:
            raise RuntimeError("S3Manager not configured for this TaskFiles instance.")
        path = join(self._task_dir, dirname) if dirname else self._task_dir
        return await self._s3_manager._upload_to_s3(path, s3_prefix=self._job_id)

    async def download_file(self, uri: str, filename: str, verify_meta: FileMetadata | None = None) -> str:
        """Downloads a file from S3 to the task directory with optional integrity check."""
        if not self._s3_manager:
            raise RuntimeError("S3Manager not configured for this TaskFiles instance.")
        return await self._s3_manager._process_s3_uri(uri, self._task_id, verify_meta=verify_meta)

    async def list(self) -> list[str]:
        """
        Asynchronously lists all file and directory names within the task root.
        """
        root = await self.get_root()
        return await listdir(root)

    async def exists(self, filename: str) -> bool:
        """
        Asynchronously checks if a specific file or directory exists in the task root.
        """
        path = join(self._task_dir, filename)
        return await aio_exists(path)

    def __repr__(self):
        return f"<TaskFiles root='{self._task_dir}'>"



from __future__ import annotations

from _socket import gaierror, gethostbyname, gethostname
from contextlib import suppress
from json import JSONDecodeError, loads
from logging import getLogger
from os import environ, getenv
from typing import Any, cast
from uuid import uuid4

from rxon.validators import validate_identifier

logger = getLogger(__name__)


class WorkerConfig:
    """A class for centralized management of worker configuration.
    Reads parameters from environment variables and provides default values.
    """

    def __init__(self) -> None:
        # --- Basic worker information ---
        self.WORKER_ID: str = getenv("WORKER_ID", f"worker-{uuid4()}")
        self.WORKER_TYPE: str = getenv("WORKER_TYPE", "generic-cpu-worker")
        self.WORKER_PORT: int = int(getenv("WORKER_PORT", "8083"))
        self.HOSTNAME: str = gethostname()
        self.IP_ADDRESS: str = "127.0.0.1"
        with suppress(gaierror):
            self.IP_ADDRESS = gethostbyname(self.HOSTNAME)

        # --- Orchestrator settings ---
        self.ORCHESTRATORS: list[dict[str, Any]] = self._get_orchestrators_config()

        # --- Security ---
        self.WORKER_TOKEN: str = getenv(
            "WORKER_INDIVIDUAL_TOKEN",
            getenv("WORKER_TOKEN", "your-secret-worker-token"),
        )
        self.TLS_CA_PATH: str | None = getenv("TLS_CA_PATH")
        self.TLS_CERT_PATH: str | None = getenv("TLS_CERT_PATH")
        self.TLS_KEY_PATH: str | None = getenv("TLS_KEY_PATH")

        # --- Resources and performance ---
        self.COST_PER_SKILL: dict[str, float] = self._load_json_from_env("COST_PER_SKILL", default={})
        self.MAX_CONCURRENT_TASKS: int = int(getenv("MAX_CONCURRENT_TASKS", "10"))
        self.RESOURCES: dict[str, Any] = {
            "cpu_cores": int(getenv("CPU_CORES", "4")),
            "gpu_info": self._get_gpu_info(),
        }

        # --- Installed software and models (read as JSON strings) ---
        self.INSTALLED_SOFTWARE: dict[str, str] = self._load_json_from_env(
            "INSTALLED_SOFTWARE",
            default={"python": "3.11"},
        )
        self.INSTALLED_MODELS: list[dict[str, str]] = self._load_json_from_env(
            "INSTALLED_MODELS",
            default=[],
        )

        # --- S3 Settings for payload offloading ---
        self.TASK_FILES_DIR: str = getenv("TASK_FILES_DIR", "/tmp/payloads")
        self.S3_ENDPOINT_URL: str | None = getenv("S3_ENDPOINT_URL")
        self.S3_ACCESS_KEY: str | None = getenv("S3_ACCESS_KEY")
        self.S3_SECRET_KEY: str | None = getenv("S3_SECRET_KEY")
        self.S3_DEFAULT_BUCKET: str = getenv("S3_DEFAULT_BUCKET", "avtomatika-payloads")
        self.S3_REGION: str = getenv("S3_REGION", "us-east-1")

        # --- Tuning parameters ---
        self.HEARTBEAT_INTERVAL: float = float(getenv("HEARTBEAT_INTERVAL", "15"))
        self.RESULT_MAX_RETRIES: int = int(getenv("RESULT_MAX_RETRIES", "5"))
        self.RESULT_RETRY_INITIAL_DELAY: float = float(
            getenv("RESULT_RETRY_INITIAL_DELAY", "1.0"),
        )
        self.HEARTBEAT_DEBOUNCE_DELAY: float = float(getenv("WORKER_HEARTBEAT_DEBOUNCE_DELAY", 0.1))
        self.TASK_POLL_TIMEOUT: float = float(getenv("TASK_POLL_TIMEOUT", "30"))
        self.TASK_POLL_ERROR_DELAY: float = float(
            getenv("TASK_POLL_ERROR_DELAY", "5.0"),
        )
        self.IDLE_POLL_DELAY: float = float(getenv("IDLE_POLL_DELAY", "0.01"))
        self.SHUTDOWN_TIMEOUT: float = float(getenv("WORKER_SHUTDOWN_TIMEOUT", "30.0"))
        self.ENABLE_WEBSOCKETS: bool = getenv("WORKER_ENABLE_WEBSOCKETS", "false").lower() == "true"
        self.MULTI_ORCHESTRATOR_MODE: str = getenv("MULTI_ORCHESTRATOR_MODE", "FAILOVER")
        self.WORKER_SKILLS_DIR: str = getenv("WORKER_SKILLS_DIR", "skills")

        # --- Custom Extra Capabilities ---
        # Automatically pick up environment variables starting with WORKER_EXTRA_
        self.EXTRA_CAPABILITIES: dict[str, Any] = self._load_extra_from_env()

    def _load_extra_from_env(self) -> dict[str, Any]:
        """Loads all environment variables starting with WORKER_EXTRA_ into a dictionary."""
        extra = {}
        prefix = "WORKER_EXTRA_"
        for key, value in environ.items():
            if key.startswith(prefix):
                # Convert WORKER_EXTRA_REGION to region
                name = key[len(prefix) :].lower()
                # Try to parse as JSON if it looks like it, otherwise keep as string
                if value.startswith(("{", "[")):
                    try:
                        extra[name] = loads(value)
                    except (JSONDecodeError, TypeError):
                        extra[name] = value
                else:
                    extra[name] = value
        return extra

    def validate(self) -> None:
        """Validates critical configuration parameters."""
        validate_identifier(self.WORKER_ID, "WORKER_ID")
        if self.WORKER_TOKEN == "your-secret-worker-token":
            logger.warning("WORKER_TOKEN is set to the default value. Tasks might fail authentication.")

        if not self.ORCHESTRATORS:
            raise ValueError("No orchestrators configured.")

        for o in self.ORCHESTRATORS:
            if not o.get("url"):
                raise ValueError("Orchestrator configuration missing URL.")

    def _get_orchestrators_config(self) -> list[dict[str, Any]]:
        """
        Loads orchestrator configuration from the ORCHESTRATORS_CONFIG environment variable.
        For backward compatibility, if it is not set, it uses ORCHESTRATOR_URL.
        """
        if orchestrators_json := getenv("ORCHESTRATORS_CONFIG"):
            try:
                orchestrators = loads(orchestrators_json)
                if getenv("ORCHESTRATOR_URL"):
                    logger.info("Both ORCHESTRATORS_CONFIG and ORCHESTRATOR_URL are set. Using ORCHESTRATORS_CONFIG.")
                for o in orchestrators:
                    if "priority" not in o:
                        o["priority"] = 10
                    if "weight" not in o:
                        o["weight"] = 1
                orchestrators.sort(key=lambda x: (x.get("priority", 10), x.get("url")))
                return cast(list[dict[str, Any]], orchestrators)
            except JSONDecodeError:
                logger.warning("Could not decode JSON from ORCHESTRATORS_CONFIG. Falling back to default.")

        orchestrator_url = getenv("ORCHESTRATOR_URL", "http://localhost:8080")
        return [{"url": orchestrator_url, "priority": 1, "weight": 1}]

    @staticmethod
    def _get_gpu_info() -> dict[str, Any] | None:
        """Collects GPU information from environment variables.
        Returns None if GPU is not configured.
        """
        if gpu_model := getenv("GPU_MODEL"):
            return {
                "model": gpu_model,
                "vram_gb": int(getenv("GPU_VRAM_GB", "0")),
            }
        else:
            return None

    @staticmethod
    def _load_json_from_env(key: str, default: Any) -> Any:
        """Safely loads a JSON string from an environment variable."""
        if value := getenv(key):
            try:
                return loads(value)
            except JSONDecodeError:
                logger.warning("Could not decode JSON from environment variable %s.", key)
                return default
        return default



import asyncio
import sys
from unittest.mock import patch

import pytest
from rxon import Transport
from rxon.constants import ERROR_CODE_INVALID_INPUT
from rxon.constants import ERROR_CODE_PERMANENT as PERMANENT_ERROR
from rxon.testing import MockTransport

from avtomatika_worker.types import ParamValidationError
from avtomatika_worker.worker import Worker


def test_pydantic_not_installed():
    """
    Tests that the worker initializes correctly when pydantic is not installed.
    """
    with patch.dict(sys.modules, {"pydantic": None}):
        from importlib import reload

        from avtomatika_worker import worker

        reload(worker)
        assert not worker._PYDANTIC_INSTALLED
    reload(worker)


@pytest.mark.filterwarnings("ignore:coroutine 'AsyncMockMixin._execute_mock_call' was never awaited:RuntimeWarning")
def test_task_decorator_warns_on_undefined_type(mocker):
    """
    Tests that the task decorator logs a warning if a task_type is not in task_type_limits.
    """
    mocker.patch("avtomatika_worker.worker.S3Manager")
    logger_mock = mocker.patch("avtomatika_worker.worker.logger")
    worker = Worker(task_type_limits={"gpu": 1})

    @worker.task("test_task", task_type="cpu")
    def my_task(params: dict):
        pass

    logger_mock.warning.assert_called_with(
        "Task 'test_task' has a type 'cpu' which is not defined in 'task_type_limits'. "
        "No concurrency limit will be applied for this type."
    )


@pytest.mark.asyncio
async def test_worker_registration_payload(mocker):
    """Tests that the registration payload contains all expected fields."""
    mock_transport = MockTransport(worker_id="test-worker")

    worker = Worker(worker_type="custom-type", clients=[({"url": "http://test", "weight": 1}, mock_transport)])
    worker._config.WORKER_ID = "custom-id"
    worker._config.INSTALLED_MODELS = [{"name": "model1", "version": "1.0"}]
    worker._config.COST_PER_SKILL = {"task1": 1.5}

    @worker.task("task1")
    def task1(params: dict):
        pass

    await worker._register_with_all_orchestrators()

    assert len(mock_transport.registered) == 1
    registration = mock_transport.registered[0]

    assert registration.worker_id == "custom-id"
    assert registration.worker_type == "custom-type"
    assert "task1" in registration.supported_skills
    assert registration.installed_models[0].name == "model1"
    assert registration.capabilities.cost_per_skill == {"task1": 1.5}


@pytest.mark.asyncio
async def test_poll_for_tasks_handles_non_204_status(mocker):
    """Tests that _poll_for_tasks handles errors from client."""
    client = mocker.AsyncMock(spec=Transport)
    client.poll_task.return_value = None

    worker = Worker()

    @worker.task("dummy_task")
    def dummy_handler(params):
        pass

    await worker._poll_for_tasks(client)
    client.poll_task.assert_called_once()


@pytest.mark.asyncio
async def test_websocket_manager_handles_connection_error(mocker):
    """
    Tests that _listen_to_single_transport handles connection errors and retries (sleeps).
    """
    worker = Worker()
    client = mocker.AsyncMock(spec=Transport)

    # Mock listen_for_commands to raise exception
    client.listen_for_commands.side_effect = Exception("Connection failed")

    mock_sleep = mocker.patch("avtomatika_worker.worker.sleep", new_callable=mocker.AsyncMock)

    # We want to run the loop once, verify sleep is called, then exit.
    # We can side_effect sleep to raise CancelledError to exit the loop.
    mock_sleep.side_effect = asyncio.CancelledError

    with pytest.raises(asyncio.CancelledError):
        await worker._listen_to_single_transport(client)

    client.listen_for_commands.assert_called()
    mock_sleep.assert_called_with(5)


@pytest.mark.asyncio
async def test_process_task_permanent_error_on_unsupported_task(mocker):
    """
    Tests that a permanent error is returned for an unsupported task type.
    """
    client = mocker.AsyncMock(spec=Transport)
    worker = Worker()

    task_data = {
        "job_id": "j1",
        "task_id": "t1",
        "type": "unsupported_task",
        "params": {},
        "tracing_context": {},
        "client": client,
    }
    await worker._process_task(task_data)

    client.send_result.assert_called_once()
    result = client.send_result.call_args.args[0]
    assert result.status == "failure"
    assert result.error.code == PERMANENT_ERROR


@pytest.mark.asyncio
async def test_prepare_task_params_raises_validation_error_for_dataclass():
    """
    Tests that _prepare_task_params raises ParamValidationError for a dataclass with missing fields.
    """
    worker = Worker()

    from dataclasses import dataclass

    @dataclass
    class MyDataclass:
        a: int
        b: str

    @worker.task("test_task")
    async def my_handler(params: MyDataclass):
        pass

    with pytest.raises(ParamValidationError):
        worker._prepare_task_params(my_handler, {"a": 1})


@pytest.mark.asyncio
async def test_process_task_handles_param_validation_error(mocker):
    """
    Tests that _process_task sends an INVALID_INPUT_ERROR when ParamValidationError is raised.
    """
    client = mocker.AsyncMock(spec=Transport)
    worker = Worker()

    @worker.task("validation_task")
    async def my_task(params: dict, **kwargs):
        raise ParamValidationError("Invalid params")

    task_data = {
        "job_id": "j1",
        "task_id": "t1",
        "type": "validation_task",
        "params": {},
        "tracing_context": {},
        "client": client,
    }

    await worker._process_task(task_data)

    client.send_result.assert_called_once()
    result = client.send_result.call_args[0][0]
    assert result.status == "failure"
    assert result.error.code == ERROR_CODE_INVALID_INPUT


def test_run_keyboard_interrupt(mocker):
    """Tests that run() handles KeyboardInterrupt gracefully."""
    worker = Worker()
    mocker.patch.object(worker, "main", side_effect=KeyboardInterrupt)
    mock_shutdown_set = mocker.patch.object(worker._shutdown_event, "set")

    worker.run()

    mock_shutdown_set.assert_called_once()


def test_run_with_health_check_keyboard_interrupt(mocker):
    """Tests that run_with_health_check() handles KeyboardInterrupt."""
    worker = Worker()
    mocker.patch.object(worker, "main", side_effect=KeyboardInterrupt)
    mock_shutdown_set = mocker.patch.object(worker._shutdown_event, "set")

    worker.run_with_health_check()

    mock_shutdown_set.assert_called_once()



import os
import tempfile
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from obstore.store import S3Store
from rxon.models import FileMetadata

from avtomatika_worker.config import WorkerConfig
from avtomatika_worker.s3 import S3Manager


@pytest.fixture
def s3_manager():
    """Provides an S3Manager instance for testing."""
    config = WorkerConfig()
    config.S3_ENDPOINT_URL = "http://localhost:5000"
    config.S3_ACCESS_KEY = "testing"
    config.S3_SECRET_KEY = "testing"
    config.S3_DEFAULT_BUCKET = "test-bucket"
    config.TASK_FILES_DIR = tempfile.gettempdir()

    manager = S3Manager(config)
    return manager


# Helper classes for mocking obstore


class MockGetResult:
    def __init__(self, data=b"test content"):
        self.data = data
        self.meta = MagicMock(size=len(data), e_tag='"hash"')

    async def stream(self):
        yield self.data


class MockObjectMeta:
    def __init__(self, key):
        self.key = key


class MockAsyncIterator:
    def __init__(self, items):
        self.items = items

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.items:
            raise StopAsyncIteration
        return self.items.pop(0)


@pytest.mark.asyncio
async def test_get_store_caching(s3_manager):
    """Tests that _get_store caches S3Store instances."""
    # We want S3Store() to return a new mock each time it's called as a constructor
    with patch(
        "avtomatika_worker.s3.S3Store", side_effect=lambda *args, **kwargs: MagicMock(spec=S3Store)
    ) as mock_s3_store:
        store1 = s3_manager._get_store("bucket1")
        store2 = s3_manager._get_store("bucket1")
        store3 = s3_manager._get_store("bucket2")

        assert store1 is store2
        assert store1 is not store3

        assert mock_s3_store.call_count == 2


@pytest.mark.asyncio
async def test_process_s3_uri_file(s3_manager):
    """Tests that _process_s3_uri downloads a single file from S3."""
    task_id = "task-123"

    with patch("avtomatika_worker.s3.obstore_get", new_callable=AsyncMock) as mock_get:
        mock_get.return_value = MockGetResult()

        with patch.object(s3_manager, "_get_store", return_value=MagicMock()) as mock_get_store:
            local_path = await s3_manager._process_s3_uri("s3://test-bucket/test-file.txt", task_id)

            # Path should include task_id
            expected_suffix = os.path.join(task_id, "test-file.txt")
            assert local_path.endswith(expected_suffix)

            mock_get_store.assert_called_with("test-bucket")
            mock_get.assert_awaited_once()

            # Check content
            with open(local_path, "rb") as f:
                assert f.read() == b"test content"


@pytest.mark.asyncio
async def test_process_s3_uri_folder(s3_manager):
    """Tests that _process_s3_uri downloads a folder (prefix) from S3."""
    task_id = "task-folder"

    mock_files = [MockObjectMeta("data/file1.txt"), MockObjectMeta("data/sub/file2.txt")]

    # obstore.list is called as: async for obj in obstore.list(...)
    # So obstore.list itself should return the async iterator.
    with (
        patch("avtomatika_worker.s3.obstore_list", return_value=MockAsyncIterator(mock_files)) as mock_list,
        patch("avtomatika_worker.s3.obstore_get", new_callable=AsyncMock) as mock_get,
    ):
        mock_get.return_value = MockGetResult()

        with patch.object(s3_manager, "_get_store", return_value=MagicMock()):
            local_path = await s3_manager._process_s3_uri("s3://test-bucket/data/", task_id)

    assert local_path.endswith(os.path.join(task_id, "data"))

    # Verify list called with correct prefix
    mock_list.assert_called_once()
    assert mock_list.call_args[1]["prefix"] == "data/"

    # Verify download calls
    assert mock_get.await_count == 2


@pytest.mark.asyncio
async def test_upload_to_s3_file(s3_manager):
    """Tests that _upload_to_s3 uploads a single file to S3."""
    local_path = os.path.join(s3_manager._config.TASK_FILES_DIR, "test-upload.txt")
    with open(local_path, "w") as f:
        f.write("test content")

    try:
        with (
            patch("avtomatika_worker.s3.obstore_put", new_callable=AsyncMock) as mock_put,
            patch.object(s3_manager, "_get_store", return_value=MagicMock()),
        ):
            mock_put.return_value = MagicMock(e_tag='"etag"')
            meta = await s3_manager._upload_to_s3(local_path)

        assert meta.uri == "s3://test-bucket/test-upload.txt"
        assert meta.etag == "etag"
        mock_put.assert_awaited_once()
        args = mock_put.await_args[0]
        assert args[1] == "test-upload.txt"  # key
        # Check that a file object was passed
        assert hasattr(args[2], "read")
    finally:
        if os.path.exists(local_path):
            os.remove(local_path)


@pytest.mark.asyncio
async def test_upload_to_s3_folder(s3_manager):
    """Tests that _upload_to_s3 uploads a folder recursively to S3."""
    # Create temp directory structure
    with tempfile.TemporaryDirectory(dir=s3_manager._config.TASK_FILES_DIR) as tmpdir:
        # Create file1
        with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
            f.write("content1")

        dirname = os.path.basename(tmpdir)

        with (
            patch("avtomatika_worker.s3.obstore_put", new_callable=AsyncMock) as mock_put,
            patch.object(s3_manager, "_get_store", return_value=MagicMock()),
        ):
            meta = await s3_manager._upload_to_s3(tmpdir)

        assert meta.uri == f"s3://test-bucket/{dirname}/"
        assert mock_put.await_count == 1


@pytest.mark.asyncio
async def test_process_params(s3_manager):
    """Tests that process_params correctly calls _process_s3_uri with task_id."""
    with patch.object(
        s3_manager,
        "_process_s3_uri",
        side_effect=lambda uri, tid, verify_meta=None: f"/local/{tid}/{uri.split('/')[-1]}",
    ):
        params = {"file": "s3://test-bucket/test-file.txt", "other": "value"}
        processed_params = await s3_manager.process_params(params, "task-1")

        assert processed_params["other"] == "value"
        assert processed_params["file"] == "/local/task-1/test-file.txt"


@pytest.mark.asyncio
async def test_process_result(s3_manager):
    """Tests that process_result correctly calls _upload_to_s3."""
    local_path = os.path.join(s3_manager._config.TASK_FILES_DIR, "output.txt")
    with open(local_path, "w") as f:
        f.write("test content")

    try:
        mock_meta = FileMetadata(uri="s3://bucket/output.txt", size=10, etag="hash")
        with patch.object(s3_manager, "_upload_to_s3", return_value=mock_meta):
            result = {"data": {"output_file": local_path}}
            processed_result, metadata = await s3_manager.process_result(result)

            assert processed_result["data"]["output_file"] == "s3://bucket/output.txt"
            assert metadata["data.output_file"] == mock_meta
    finally:
        if os.path.exists(local_path):
            os.remove(local_path)


@pytest.mark.asyncio
async def test_cleanup(s3_manager):
    """Tests that cleanup removes the task directory."""
    task_id = "cleanup-task"
    task_dir = os.path.join(s3_manager._config.TASK_FILES_DIR, task_id)
    os.makedirs(task_dir, exist_ok=True)
    with open(os.path.join(task_dir, "file.txt"), "w") as f:
        f.write("data")

    assert os.path.exists(task_dir)

    await s3_manager.cleanup(task_id)

    assert not os.path.exists(task_dir)



import io
import json
import logging

import pytest
from rxon.testing import MockTransport

from avtomatika_worker.logging import _context_filter, clear_context, set_context, setup_logging
from avtomatika_worker.worker import Worker


def test_setup_logging_text(monkeypatch):
    monkeypatch.setenv("LOG_FORMAT", "text")
    setup_logging(worker_id="test-worker")

    root_logger = logging.getLogger()
    handler = root_logger.handlers[0]
    assert isinstance(handler.formatter, logging.Formatter)


def test_context_filter():
    _context_filter.context = {}
    set_context(test_key="test_value")

    record = logging.LogRecord("name", logging.INFO, "pathname", 10, "message", None, None)
    _context_filter.filter(record)

    assert record.test_key == "test_value"

    clear_context("test_key")
    record2 = logging.LogRecord("name", logging.INFO, "pathname", 10, "message", None, None)
    _context_filter.filter(record2)
    assert not hasattr(record2, "test_key")


@pytest.mark.asyncio
async def test_worker_logging_context(monkeypatch):
    monkeypatch.setenv("LOG_FORMAT", "json")

    # Clear loggers before test
    root_logger = logging.getLogger()
    for h in root_logger.handlers[:]:
        root_logger.removeHandler(h)

    transport = MockTransport(worker_id="worker-1")
    # Capture output via StreamHandler or check record in ContextFilter

    worker = Worker(worker_type="test", clients=[({"url": "http://test", "weight": 1}, transport)])

    # Setup log capture after worker initialization
    log_capture = io.StringIO()
    handler = logging.StreamHandler(log_capture)
    # Use the same format as in setup_logging
    from pythonjsonlogger import json as jsonlogger

    formatter = jsonlogger.JsonFormatter(
        "%(asctime)s %(levelname)s %(name)s %(message)s %(worker_id)s %(task_id)s %(job_id)s"
    )
    handler.setFormatter(formatter)
    handler.addFilter(_context_filter)
    root_logger.addHandler(handler)

    @worker.task("test_task")
    async def task_handler(params, **kwargs):
        logging.getLogger("test").info("Inside task")
        return {"status": "success"}

    task_data = {
        "job_id": "job-123",
        "task_id": "task-456",
        "type": "test_task",
        "params": {},
        "tracing_context": {},
        "client": transport,
    }

    await worker._process_task(task_data)

    log_output = log_capture.getvalue().splitlines()
    # Find our log "Inside task"
    task_log = next(line for line in log_output if "Inside task" in line)
    log_data = json.loads(task_log)

    assert log_data["task_id"] == "task-456"
    assert log_data["job_id"] == "job-123"
    assert log_data["message"] == "Inside task"



from avtomatika_worker import types


def test_types_constants():
    """Tests that the constants in the types module are defined correctly."""
    assert types.TRANSIENT_ERROR == "TRANSIENT_ERROR"
    assert types.PERMANENT_ERROR == "PERMANENT_ERROR"
    assert types.INVALID_INPUT_ERROR == "INVALID_INPUT_ERROR"



import os

import pytest
from rxon import Transport

from avtomatika_worker import Worker
from avtomatika_worker.task_files import TaskFiles


@pytest.mark.asyncio
async def test_task_files_injection(mocker):
    """Tests that TaskFiles object is correctly injected into the handler."""
    client = mocker.AsyncMock(spec=Transport)

    worker = Worker()
    worker._s3_manager.process_params = mocker.AsyncMock(return_value={})
    worker._s3_manager.process_result = mocker.AsyncMock(return_value={})
    worker._s3_manager.cleanup = mocker.AsyncMock()

    injected_task_files = None

    @worker.task("di_task")
    async def handler(params: dict, tf: TaskFiles, **kwargs):
        nonlocal injected_task_files
        injected_task_files = tf
        # Check async usage
        await tf.path_to("test.txt")
        return {"status": "success"}

    task_data = {
        "job_id": "j1",
        "task_id": "t1",
        "type": "di_task",
        "params": {},
        "tracing_context": {},
        "client": client,
    }

    await worker._process_task(task_data)

    assert isinstance(injected_task_files, TaskFiles)
    # Check if root is correct
    expected_root = os.path.join(worker._config.TASK_FILES_DIR, "t1")
    # Verify the path is correct
    assert injected_task_files._task_dir == expected_root

    # Verify cleanup was called
    worker._s3_manager.cleanup.assert_awaited_once_with("t1")


@pytest.mark.asyncio
async def test_task_files_read_write(tmp_path):
    """Tests high-level read/write methods of TaskFiles."""
    target_dir = tmp_path / "task_dir"
    tf = TaskFiles(str(target_dir), job_id="job1", task_id="task1")

    filename = "hello.txt"
    content = "Hello, World!"

    await tf.write(filename, content)
    assert (target_dir / filename).exists()

    read_content = await tf.read(filename)
    assert read_content == content


@pytest.mark.asyncio
async def test_task_files_list_exists(tmp_path):
    """Tests list() and exists() methods."""
    target_dir = tmp_path / "task_dir"
    tf = TaskFiles(str(target_dir), job_id="job1", task_id="task1")

    await tf.write("file1.txt", "data1")
    await tf.write("file2.txt", "data2")

    files = await tf.list()
    assert sorted(files) == ["file1.txt", "file2.txt"]

    assert await tf.exists("file1.txt")
    assert not await tf.exists("missing.txt")


@pytest.mark.asyncio
async def test_task_files_open_context(tmp_path):
    """Tests the async with files.open(...) context manager."""
    target_dir = tmp_path / "task_dir"
    tf = TaskFiles(str(target_dir), job_id="job1", task_id="task1")

    async with tf.open("stream.txt", mode="w") as f:
        await f.write("line1\n")
        await f.write("line2\n")

    content = await tf.read("stream.txt")
    assert content == "line1\nline2\n"


@pytest.mark.asyncio
async def test_task_files_nested_creation(tmp_path):
    """Tests that nested directories are created automatically."""
    target_dir = tmp_path / "task_dir"
    tf = TaskFiles(str(target_dir), job_id="job1", task_id="task1")

    # Path with subdirectories
    nested_path = "sub/dir/test.txt"
    await tf.write(nested_path, "nested")

    assert (target_dir / "sub" / "dir" / "test.txt").exists()
    assert await tf.read(nested_path) == "nested"


@pytest.mark.asyncio
async def test_task_files_path_to(tmp_path):
    """Tests path_to method."""
    target_dir = tmp_path / "task_dir"
    tf = TaskFiles(str(target_dir), job_id="job1", task_id="task1")

    file_path = await tf.path_to("output.txt")

    assert target_dir.exists()
    assert file_path == os.path.join(str(target_dir), "output.txt")



from avtomatika_worker.config import WorkerConfig
from avtomatika_worker.worker import Worker


def test_task_registration_with_type():
    """Tests that the @worker.task decorator correctly registers a task with a type."""
    worker = Worker(task_type_limits={"video": 1})

    @worker.task("process_video", task_type="video")
    async def my_handler(params: dict):
        return {"status": "success"}

    assert "process_video" in worker._task_handlers
    assert worker._task_handlers["process_video"]["type"] == "video"
    assert worker._task_handlers["process_video"]["func"] == my_handler


def test_get_current_state_initial():
    """Tests the initial state of the worker."""
    mock_config = WorkerConfig()
    mock_config.MAX_CONCURRENT_TASKS = 5  # Set initial limit
    mock_config.ORCHESTRATORS = [{"url": "http://test", "weight": 1}]  # Needs to be defined

    worker = Worker(config=mock_config, task_type_limits={"video": 1, "audio": 2})

    @worker.task("process_video", task_type="video")
    async def video_handler(params: dict): ...

    @worker.task("process_audio", task_type="audio")
    async def audio_handler(params: dict): ...

    @worker.task("unlimited_task")
    async def unlimited_handler(params: dict): ...

    state = worker._get_current_state()
    assert state["status"] == "idle"
    assert set(state["supported_skills"]) == {
        "process_video",
        "process_audio",
        "unlimited_task",
    }


def test_get_current_state_global_limit_reached():
    """Tests that the worker becomes busy when the global concurrency limit is reached."""
    # Create a mock config to ensure MAX_CONCURRENT_TASKS is controlled
    mock_config = WorkerConfig()
    mock_config.MAX_CONCURRENT_TASKS = 1

    worker = Worker(config=mock_config)  # Pass the mock config
    worker._current_load = 1

    @worker.task("some_task")
    async def some_task(params: dict): ...

    state = worker._get_current_state()
    assert state["status"] == "busy"
    assert state["supported_skills"] == []


def test_get_current_state_type_limit_reached():
    """Tests that a task type is unavailable when its limit is reached."""
    worker = Worker(max_concurrent_tasks=5, task_type_limits={"video": 1})
    worker._current_load = 1
    worker._current_load_by_type["video"] = 1

    @worker.task("process_video", task_type="video")
    async def video_handler(params: dict): ...

    @worker.task("process_audio", task_type="audio")
    async def audio_handler(params: dict): ...

    state = worker._get_current_state()
    assert state["status"] == "idle"
    assert set(state["supported_skills"]) == {"process_audio"}


def test_get_current_state_all_type_limits_reached():
    """Tests that the worker becomes busy when all type limits are reached,
    even if the global limit is not."""
    worker = Worker(max_concurrent_tasks=5, task_type_limits={"video": 1, "audio": 1})
    worker._current_load = 2
    worker._current_load_by_type["video"] = 1
    worker._current_load_by_type["audio"] = 1

    @worker.task("process_video", task_type="video")
    async def video_handler(params: dict): ...

    @worker.task("process_audio", task_type="audio")
    async def audio_handler(params: dict): ...

    state = worker._get_current_state()
    assert state["status"] == "busy"
    assert state["supported_skills"] == []



import pytest

from avtomatika_worker.config import WorkerConfig
from avtomatika_worker.worker import Worker


def test_config_validation_invalid_worker_id(monkeypatch):
    """Test that WorkerConfig raises ValueError when WORKER_ID contains invalid characters."""
    monkeypatch.setenv("WORKER_ID", "invalid/id")
    monkeypatch.setenv("ORCHESTRATOR_URL", "http://localhost:8080")
    # Config init reads env vars immediately
    config = WorkerConfig()

    with pytest.raises(ValueError, match="Invalid WORKER_ID"):
        config.validate()


def test_config_validation_valid_worker_id(monkeypatch):
    """Test that WorkerConfig passes validation with a valid WORKER_ID."""
    monkeypatch.setenv("WORKER_ID", "valid-worker_123")
    monkeypatch.setenv("ORCHESTRATOR_URL", "http://localhost:8080")
    config = WorkerConfig()
    # Should not raise
    config.validate()


def test_task_decorator_validation_invalid_name(monkeypatch):
    monkeypatch.setenv("ORCHESTRATOR_URL", "http://localhost:8080")
    worker = Worker()

    with pytest.raises(ValueError, match="Invalid task name"):

        @worker.task(name="invalid/task name")
        async def my_task(params, **kwargs):
            pass


def test_task_decorator_validation_invalid_type(monkeypatch):
    monkeypatch.setenv("ORCHESTRATOR_URL", "http://localhost:8080")
    worker = Worker()

    with pytest.raises(ValueError, match="Invalid task type"):

        @worker.task(name="valid_task", task_type="invalid type!")
        async def my_task(params, **kwargs):
            pass


def test_task_decorator_validation_success(monkeypatch):
    monkeypatch.setenv("ORCHESTRATOR_URL", "http://localhost:8080")
    worker = Worker()

    @worker.task(name="valid_task", task_type="valid_type")
    async def my_task(params, **kwargs):
        pass

    assert "valid_task" in worker._task_handlers



"""A Python SDK for creating workers for the Avtomatika Orchestrator."""

from importlib.metadata import PackageNotFoundError, version

from .logging import setup_logging
from .task_files import TaskFiles
from .worker import SkillBlueprint, Worker

__all__ = ["Worker", "SkillBlueprint", "TaskFiles", "setup_logging"]

try:
    __version__ = version("avtomatika-worker")
except PackageNotFoundError:
    __version__ = "unknown"



from __future__ import annotations

from asyncio import (
    CancelledError,
    Event,
    Task,
    TaskGroup,
    create_task,
    gather,
    get_running_loop,
    run,
    sleep,
    to_thread,
    wait_for,
)
from contextlib import suppress
from dataclasses import fields, is_dataclass
from importlib.util import module_from_spec, spec_from_file_location
from inspect import Parameter, iscoroutinefunction, signature
from logging import getLogger
from os.path import join
from signal import SIGINT, SIGTERM
from typing import Any, Callable, cast, get_args, get_origin, get_type_hints

from aiofiles.os import listdir
from aiofiles.ospath import abspath, exists
from aiohttp import ClientSession, TCPConnector, web
from rxon import Transport, create_transport
from rxon.blob import calculate_config_hash
from rxon.constants import (
    COMMAND_CANCEL_TASK,
    ERROR_CODE_INTEGRITY_MISMATCH,
    ERROR_CODE_INVALID_INPUT,
    ERROR_CODE_PERMANENT,
    ERROR_CODE_TRANSIENT,
    MSG_TYPE_PROGRESS,
    TASK_STATUS_CANCELLED,
    TASK_STATUS_FAILURE,
    TASK_STATUS_SUCCESS,
)
from rxon.exceptions import RxonError
from rxon.models import (
    FileMetadata,
    GPUInfo,
    Heartbeat,
    InstalledModel,
    ProgressUpdatePayload,
    Resources,
    TaskError,
    TaskPayload,
    TaskResult,
    WorkerCapabilities,
    WorkerRegistration,
)
from rxon.security import create_client_ssl_context
from rxon.utils import to_dict
from rxon.validators import validate_identifier

from .config import WorkerConfig
from .logging import clear_context, set_context, setup_logging
from .s3 import S3Manager
from .task_files import TaskFiles
from .types import CapacityChecker, Middleware, ParamValidationError

try:
    from pydantic import BaseModel, ValidationError

    _PYDANTIC_INSTALLED = True
except ImportError:
    _PYDANTIC_INSTALLED = False

# Logging setup
logger = getLogger(__name__)


class SkillBlueprint:
    """
    A collection of tasks that can be registered on a Worker.
    Allows defining tasks in separate files without a worker instance.
    """

    def __init__(self) -> None:
        self._tasks: list[tuple[str, str | None, Callable]] = []

    def task(self, name: str, task_type: str | None = None) -> Callable:
        """Decorator to register a task handler on the blueprint."""

        def decorator(func: Callable) -> Callable:
            self._tasks.append((name, task_type, func))
            return func

        return decorator


class Worker:
    """The main class for creating and running a worker.
    Implements a hybrid interaction model with the Orchestrator:
    - PULL model for fetching tasks.
    - Transport layer for real-time commands (cancellation) and sending progress.
    """

    def __init__(
        self,
        worker_type: str = "generic-worker",
        max_concurrent_tasks: int | None = None,
        task_type_limits: dict[str, int] | None = None,
        http_session: ClientSession | None = None,
        skill_dependencies: dict[str, list[str]] | None = None,
        config: WorkerConfig | None = None,
        capacity_checker: CapacityChecker | None = None,
        clients: list[tuple[dict[str, Any], Transport]] | None = None,
        skills_dir: str | None = None,
    ):
        self._config = config or WorkerConfig()
        if skills_dir:
            self._config.WORKER_SKILLS_DIR = skills_dir

        setup_logging(worker_id=self._config.WORKER_ID)
        self._s3_manager = S3Manager(self._config)
        self._config.WORKER_TYPE = worker_type  # Allow overriding worker_type
        if max_concurrent_tasks is not None:
            self._config.MAX_CONCURRENT_TASKS = max_concurrent_tasks

        self._task_type_limits = task_type_limits or {}
        self._task_handlers: dict[str, dict[str, Any]] = {}
        self._skill_dependencies = skill_dependencies or {}
        self._middlewares: list[Middleware] = []
        self._capacity_checker = capacity_checker

        # Worker state
        self._current_load = 0
        self._current_load_by_type: dict[str, int] = dict.fromkeys(self._task_type_limits, 0)
        self._hot_cache: set[str] = set()
        self._active_tasks: dict[str, Task] = {}
        self._http_session = http_session
        self._session_is_managed_externally = http_session is not None
        self._shutdown_event = Event()
        self._registered_event = Event()
        self._debounce_task: Task | None = None
        self._ssl_context = None

        # --- Weighted Round-Robin State ---
        self._total_orchestrator_weight = 0
        if self._config.ORCHESTRATORS:
            for o in self._config.ORCHESTRATORS:
                o["current_weight"] = 0
                self._total_orchestrator_weight += o.get("weight", 1)
        self._clients = clients or []
        if not self._clients and self._http_session:
            self._init_clients()

    def add_middleware(self, middleware: Middleware) -> None:
        """Adds a middleware to the execution chain."""
        self._middlewares.append(middleware)

    def include_blueprint(self, blueprint: SkillBlueprint) -> None:
        """Registers all tasks from a blueprint."""
        for name, task_type, func in blueprint._tasks:
            self.task(name, task_type)(func)

    async def load_skills(self, skills_dir: str | None = None) -> None:
        """
        Dynamically loads skills from the specified directory.
        Scans for .py files and looks for 'SkillBlueprint' instances or 'setup(worker)' functions.
        """
        skills_dir = skills_dir or self._config.WORKER_SKILLS_DIR
        if not await exists(skills_dir):
            return

        logger.info(f"Scanning for skills in: {await abspath(skills_dir)}")
        for filename in await listdir(skills_dir):
            if filename.endswith(".py") and not filename.startswith("_"):
                module_name = filename[:-3]
                file_path = join(skills_dir, filename)

                try:
                    spec = spec_from_file_location(module_name, file_path)
                    if spec is None or spec.loader is None:
                        continue
                    module = module_from_spec(spec)
                    spec.loader.exec_module(module)

                    # 1. Look for SkillBlueprint instances
                    found_anything = False
                    for attr_name in dir(module):
                        attr = getattr(module, attr_name)
                        if isinstance(attr, SkillBlueprint):
                            self.include_blueprint(attr)
                            found_anything = True

                    # 2. Look for setup(worker) function
                    if hasattr(module, "setup") and callable(module.setup):
                        module.setup(self)
                        found_anything = True

                    if found_anything:
                        logger.info(f"Loaded skills from: {filename}")
                except Exception as e:
                    logger.error(f"Failed to load skills from {filename}: {e}")

    def _init_clients(self) -> None:
        """Initializes Transport instances for each configured orchestrator."""
        session_to_use = self._http_session if self._session_is_managed_externally else None
        self._clients = [
            (
                o,
                create_transport(
                    url=o["url"],
                    worker_id=self._config.WORKER_ID,
                    token=o.get("token", self._config.WORKER_TOKEN),
                    ssl_context=self._ssl_context,
                    session=session_to_use,
                    result_retries=self._config.RESULT_MAX_RETRIES,
                    result_retry_delay=self._config.RESULT_RETRY_INITIAL_DELAY,
                ),
            )
            for o in self._config.ORCHESTRATORS
        ]

    def _validate_task_types(self) -> None:
        """Checks for unused task type limits and warns the user."""
        registered_task_types = {
            handler_data["type"] for handler_data in self._task_handlers.values() if handler_data["type"]
        }

        for task_type in self._task_type_limits:
            if task_type not in registered_task_types:
                logger.warning(
                    f"Configuration warning: A limit is defined for task type '{task_type}', "
                    "but no tasks are registered with this type."
                )

    def task(self, name: str, task_type: str | None = None) -> Callable:
        """Decorator to register a function as a task handler."""
        validate_identifier(name, "task name")
        if task_type:
            validate_identifier(task_type, "task type")

        def decorator(func: Callable) -> Callable:
            logger.info(f"Registering task: '{name}' (type: {task_type or 'N/A'})")
            if task_type and task_type not in self._task_type_limits:
                logger.warning(
                    f"Task '{name}' has a type '{task_type}' which is not defined in 'task_type_limits'. "
                    "No concurrency limit will be applied for this type."
                )
            if task_type and task_type not in self._current_load_by_type:
                self._current_load_by_type[task_type] = 0
            self._task_handlers[name] = {"func": func, "type": task_type}
            return func

        return decorator

    def add_to_hot_cache(self, model_name: str) -> None:
        """Adds a model to the hot cache."""
        self._hot_cache.add(model_name)
        self._schedule_heartbeat_debounce()

    def remove_from_hot_cache(self, model_name: str) -> None:
        """Removes a model from the hot cache."""
        self._hot_cache.discard(model_name)
        self._schedule_heartbeat_debounce()

    def get_hot_cache(self) -> set[str]:
        """Returns the hot cache."""
        return self._hot_cache

    def _get_current_state(self) -> dict[str, Any]:
        """
        Calculates the current worker state including status and available skills.
        """
        if self._current_load >= self._config.MAX_CONCURRENT_TASKS:
            return {"status": "busy", "supported_skills": []}

        supported_skills = []
        for name, handler_data in self._task_handlers.items():
            is_available = True
            task_type = handler_data.get("type")

            if task_type and task_type in self._task_type_limits:
                limit = self._task_type_limits[task_type]
                current_load = self._current_load_by_type.get(task_type, 0)
                if current_load >= limit:
                    is_available = False

            if is_available:
                supported_skills.append(name)

        if self._capacity_checker:
            supported_skills = [skill for skill in supported_skills if self._capacity_checker(skill)]

        status = "idle" if supported_skills else "busy"
        return {"status": status, "supported_skills": supported_skills}

    def _get_next_client(self) -> Transport | None:
        """
        Selects the next orchestrator client using a smooth weighted round-robin algorithm.
        """
        if not self._clients:
            return None

        # The orchestrator with the highest current_weight is selected.
        selected_client = None
        highest_weight = -1

        for o, client in self._clients:
            o["current_weight"] += o["weight"]
            if o["current_weight"] > highest_weight:
                highest_weight = o["current_weight"]
                selected_client = client

        if selected_client:
            # Find the config for the selected client to decrement its weight
            for o, client in self._clients:
                if client == selected_client:
                    o["current_weight"] -= self._total_orchestrator_weight
                    break

        return selected_client

    async def _debounced_heartbeat_sender(self) -> None:
        """Waits for the debounce delay then sends a heartbeat."""
        await sleep(self._config.HEARTBEAT_DEBOUNCE_DELAY)
        await self._send_heartbeats_to_all()

    def _schedule_heartbeat_debounce(self) -> None:
        """Schedules a debounced heartbeat, cancelling any pending one."""
        # Cancel the previously scheduled task, if it exists and is not done.
        if self._debounce_task and not self._debounce_task.done():
            self._debounce_task.cancel()
        # Schedule the new debounced call.
        self._debounce_task = create_task(self._debounced_heartbeat_sender())

    async def _poll_for_tasks(self, client: Transport) -> None:
        """Polls a specific Orchestrator for new tasks."""
        current_state = self._get_current_state()
        if current_state["status"] == "busy":
            return

        try:
            task_data = await client.poll_task(timeout=self._config.TASK_POLL_TIMEOUT)
            if task_data:
                task_data_dict = to_dict(task_data)
                task_data_dict["client"] = client

                self._current_load += 1
                if (task_handler_info := self._task_handlers.get(task_data.type)) and (
                    task_type_for_limit := task_handler_info.get("type")
                ):
                    self._current_load_by_type[task_type_for_limit] += 1
                self._schedule_heartbeat_debounce()

                task = create_task(self._process_task(task_data_dict))
                self._active_tasks[task_data.task_id] = task
        except RxonError as e:
            logger.error(f"Error polling tasks: {e}")

    async def _start_polling(self) -> None:
        """The main loop for polling tasks."""
        await self._registered_event.wait()

        while not self._shutdown_event.is_set():
            if self._get_current_state()["status"] == "busy":
                await sleep(self._config.IDLE_POLL_DELAY)
                continue

            if self._config.MULTI_ORCHESTRATOR_MODE == "ROUND_ROBIN":
                if client := self._get_next_client():
                    await self._poll_for_tasks(client)
            else:
                for _, client in self._clients:
                    if self._get_current_state()["status"] == "busy":
                        break
                    await self._poll_for_tasks(client)

            if self._current_load == 0:
                await sleep(self._config.IDLE_POLL_DELAY)

    @staticmethod
    def _prepare_task_params(handler: Callable, params: dict[str, Any]) -> Any:
        """
        Inspects the handler's signature to validate and instantiate params.
        """
        sig = signature(handler)
        params_param = sig.parameters.get("params")
        if params_param is None:
            return params
        params_annotation = params_param.annotation

        if params_annotation is sig.empty or params_annotation is dict:
            return params

        # Pydantic Model Validation
        if _PYDANTIC_INSTALLED and isinstance(params_annotation, type) and issubclass(params_annotation, BaseModel):
            try:
                return cast(Any, params_annotation).model_validate(params)
            except ValidationError as e:
                raise ParamValidationError(str(e)) from e

        # Dataclass Instantiation
        if isinstance(params_annotation, type) and is_dataclass(params_annotation):
            try:
                # Filter unknown fields
                known_fields = {f.name for f in fields(params_annotation)}
                filtered_params = {k: v for k, v in params.items() if k in known_fields}

                # Check required fields
                required_fields = [
                    f.name
                    for f in fields(params_annotation)
                    if f.default is Parameter.empty and f.default_factory is Parameter.empty
                ]

                if missing_fields := [f for f in required_fields if f not in filtered_params]:
                    raise ParamValidationError(f"Missing required fields for dataclass: {', '.join(missing_fields)}")

                return params_annotation(**filtered_params)
            except (TypeError, ValueError) as e:
                raise ParamValidationError(str(e)) from e

        return params

    def _prepare_dependencies(self, handler: Callable, job_id: str, task_id: str) -> dict[str, Any]:
        """Injects dependencies based on type hints."""
        deps = {}
        task_dir = join(self._config.TASK_FILES_DIR, task_id)
        task_files = TaskFiles(task_dir, job_id=job_id, task_id=task_id, s3_manager=self._s3_manager)

        try:
            hints = get_type_hints(handler)
        except Exception:
            # Fallback for handlers where hints can't be resolved
            hints = {}

        for name, annotation in hints.items():
            origin = get_origin(annotation)
            types = get_args(annotation) if origin else (annotation,)

            if TaskFiles in types:
                deps[name] = task_files

        return deps

    async def _process_task(self, task_data_raw: dict[str, Any]) -> None:
        """Executes the task logic."""
        client: Transport = task_data_raw.pop("client")

        # Parse incoming task data using protocol model
        if "params_metadata" in task_data_raw and task_data_raw["params_metadata"]:
            task_data_raw["params_metadata"] = {
                k: self._from_dict(FileMetadata, v) for k, v in task_data_raw["params_metadata"].items()
            }

        task_payload = self._from_dict(TaskPayload, task_data_raw)
        task_id, job_id, task_name = task_payload.task_id, task_payload.job_id, task_payload.type
        params = task_payload.params

        set_context(task_id=task_id, job_id=job_id)

        handler_data = self._task_handlers.get(task_name)
        task_type_for_limit = handler_data.get("type") if handler_data else None

        result_obj: TaskResult | None = None

        # Create a progress sender wrapper attached to this specific client
        async def send_progress_wrapper(task_id_arg, job_id_arg, progress, message=""):
            payload = ProgressUpdatePayload(
                event=MSG_TYPE_PROGRESS, task_id=task_id_arg, job_id=job_id_arg, progress=progress, message=message
            )
            await client.send_progress(payload)

        try:
            if not handler_data:
                message = f"Unsupported task: {task_name}"
                logger.warning(message)
                error = TaskError(code=ERROR_CODE_PERMANENT, message=message)
                result_obj = TaskResult(
                    job_id=job_id,
                    task_id=task_id,
                    worker_id=self._config.WORKER_ID,
                    status=TASK_STATUS_FAILURE,
                    error=error,
                )
            else:
                # Download files
                params = await self._s3_manager.process_params(params, task_id, metadata=task_payload.params_metadata)
                validated_params = self._prepare_task_params(handler_data["func"], params)
                deps = self._prepare_dependencies(handler_data["func"], job_id, task_id)

                handler_kwargs = {
                    "params": validated_params,
                    "task_id": task_id,
                    "job_id": job_id,
                    "tracing_context": task_payload.tracing_context,
                    "priority": task_data_raw.get("priority", 0),
                    "send_progress": send_progress_wrapper,
                    "add_to_hot_cache": self.add_to_hot_cache,
                    "remove_from_hot_cache": self.remove_from_hot_cache,
                    **deps,
                }

                middleware_context = {
                    "task_id": task_id,
                    "job_id": job_id,
                    "task_name": task_name,
                    "params": validated_params,
                    "handler_kwargs": handler_kwargs,
                }

                async def _execution_logic() -> Any:
                    handler = handler_data["func"]
                    final_kwargs = middleware_context["handler_kwargs"]

                    if iscoroutinefunction(handler):
                        return await handler(**final_kwargs)
                    else:
                        return await to_thread(handler, **final_kwargs)

                handler_chain = _execution_logic
                for middleware in reversed(self._middlewares):

                    def make_wrapper(mw: Middleware, next_handler: Callable) -> Callable:
                        async def wrapper():
                            return await mw(middleware_context, next_handler)

                        return wrapper

                    handler_chain = make_wrapper(middleware, handler_chain)

                handler_result = await handler_chain()

                updated_data, metadata_map = await self._s3_manager.process_result(
                    handler_result.get("data", {}), s3_prefix=job_id
                )

                # Prepare error object safely
                task_error = None
                if "error" in handler_result:
                    err_data = handler_result["error"]
                    if isinstance(err_data, dict):
                        # Ensure we only pass valid fields to TaskError
                        valid_err_fields = {k: v for k, v in err_data.items() if k in TaskError._fields}
                        task_error = TaskError(**valid_err_fields)
                    else:
                        task_error = TaskError(code=ERROR_CODE_TRANSIENT, message=str(err_data))

                result_obj = TaskResult(
                    job_id=job_id,
                    task_id=task_id,
                    worker_id=self._config.WORKER_ID,
                    status=handler_result.get("status", TASK_STATUS_SUCCESS),
                    data=updated_data,
                    error=task_error,
                    data_metadata=metadata_map if metadata_map else None,
                )

        except ParamValidationError as e:
            logger.error(f"Task {task_id} failed validation: {e}")
            error = TaskError(code=ERROR_CODE_INVALID_INPUT, message=str(e))
            result_obj = TaskResult(
                job_id=job_id,
                task_id=task_id,
                worker_id=self._config.WORKER_ID,
                status=TASK_STATUS_FAILURE,
                error=error,
            )
        except CancelledError:
            logger.info(f"Task {task_id} was cancelled.")
            result_obj = TaskResult(
                job_id=job_id, task_id=task_id, worker_id=self._config.WORKER_ID, status=TASK_STATUS_CANCELLED
            )
            raise
        except ValueError as e:
            logger.error(f"Data integrity or validation error for task {task_id}: {e}")
            error = TaskError(code=ERROR_CODE_INTEGRITY_MISMATCH, message=str(e))
            result_obj = TaskResult(
                job_id=job_id,
                task_id=task_id,
                worker_id=self._config.WORKER_ID,
                status=TASK_STATUS_FAILURE,
                error=error,
            )
        except Exception as e:
            logger.exception(f"An unexpected error occurred while processing task {task_id}:")
            error = TaskError(code=ERROR_CODE_TRANSIENT, message=str(e))
            result_obj = TaskResult(
                job_id=job_id,
                task_id=task_id,
                worker_id=self._config.WORKER_ID,
                status=TASK_STATUS_FAILURE,
                error=error,
            )
        finally:
            await self._s3_manager.cleanup(task_id)

            if result_obj:
                try:
                    accepted = await client.send_result(result_obj)
                    if not accepted:
                        logger.warning(
                            f"Task {task_id} result was IGNORED by orchestrator (possibly deadline exceeded)."
                        )
                except RxonError as e:
                    logger.error(f"Failed to send task result: {e}")

            clear_context("task_id", "job_id")
            self._active_tasks.pop(task_id, None)
            self._current_load -= 1
            if task_type_for_limit:
                self._current_load_by_type[task_type_for_limit] -= 1
            self._schedule_heartbeat_debounce()

    async def _manage_orchestrator_communications(self) -> None:
        """Registers the worker and sends heartbeats."""
        await self._register_with_all_orchestrators()

        self._registered_event.set()
        if self._config.ENABLE_WEBSOCKETS:
            create_task(self._start_websocket_manager())

        while not self._shutdown_event.is_set():
            await self._send_heartbeats_to_all()
            await sleep(self._config.HEARTBEAT_INTERVAL)

    @staticmethod
    def _from_dict(cls: type, data: Any) -> Any:
        """Safely instantiates a NamedTuple from a dict, ignoring unknown fields."""
        if not data:
            return None
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            return data
        fields = cast(Any, cls)._fields
        filtered_data = {k: v for k, v in data.items() if k in fields}
        return cls(**filtered_data)

    async def _register_with_all_orchestrators(self) -> None:
        """Registers the worker with all orchestrators."""
        state = self._get_current_state()

        gpu_info = None
        if self._config.RESOURCES.get("gpu_info"):
            gpu_info = GPUInfo(**self._config.RESOURCES["gpu_info"])

        resources = Resources(
            max_concurrent_tasks=self._config.MAX_CONCURRENT_TASKS,
            cpu_cores=self._config.RESOURCES["cpu_cores"],
            gpu_info=gpu_info,
        )

        s3_hash = calculate_config_hash(
            self._config.S3_ENDPOINT_URL,
            self._config.S3_ACCESS_KEY,
            self._config.S3_DEFAULT_BUCKET,
        )

        # Merge default extra fields with dynamic ones from ENV
        combined_extra = {"websockets": self._config.ENABLE_WEBSOCKETS}
        combined_extra.update(self._config.EXTRA_CAPABILITIES)

        registration = WorkerRegistration(
            worker_id=self._config.WORKER_ID,
            worker_type=self._config.WORKER_TYPE,
            supported_skills=state["supported_skills"],
            resources=resources,
            installed_software=self._config.INSTALLED_SOFTWARE,
            installed_models=[InstalledModel(**m) for m in self._config.INSTALLED_MODELS],
            capabilities=WorkerCapabilities(
                hostname=self._config.HOSTNAME,
                ip_address=self._config.IP_ADDRESS,
                cost_per_skill=self._config.COST_PER_SKILL,
                s3_config_hash=s3_hash,
                extra=combined_extra,
            ),
        )

        await gather(*[self._safe_register(client, registration) for _, client in self._clients])

    @staticmethod
    async def _safe_register(client: Transport, registration: WorkerRegistration) -> None:
        try:
            resp = await client.register(registration)
            if isinstance(resp, dict) and (warning := resp.get("warning")):
                logger.warning(f"Registration warning from {client}: {warning}")
        except RxonError as e:
            logger.error(f"Registration failed for {client}: {e}")

    async def _send_heartbeats_to_all(self) -> None:
        """Sends heartbeat messages to all orchestrators."""
        state = self._get_current_state()

        hot_skills = None
        if self._skill_dependencies:
            hot_skills = [
                skill for skill, models in self._skill_dependencies.items() if set(models).issubset(self._hot_cache)
            ]

        heartbeat = Heartbeat(
            worker_id=self._config.WORKER_ID,
            status=state["status"],
            load=float(self._current_load),
            current_tasks=list(self._active_tasks.keys()),
            supported_skills=state["supported_skills"],
            hot_cache=list(self._hot_cache),
            skill_dependencies=self._skill_dependencies or None,
            hot_skills=hot_skills or None,
        )

        await gather(*[self._safe_heartbeat(client, heartbeat) for _, client in self._clients])

    async def _safe_heartbeat(self, client: Transport, heartbeat: Heartbeat) -> None:
        try:
            resp = await client.send_heartbeat(heartbeat)
            if resp and "cancel_task_ids" in resp:
                for task_id in resp["cancel_task_ids"]:
                    if task_id in self._active_tasks:
                        logger.warning(f"Task {task_id} marked for cancellation via heartbeat feedback loop.")
                        self._active_tasks[task_id].cancel()
        except RxonError as e:
            logger.warning(f"Heartbeat failed for {client}: {e}")

    async def main(self) -> None:
        """The main asynchronous function."""
        self._config.validate()
        self._validate_task_types()

        await self.load_skills()

        if self._config.EXTRA_CAPABILITIES:
            logger.info(f"Loaded custom capabilities from environment: {self._config.EXTRA_CAPABILITIES}")

        loop = get_running_loop()
        for sig in (SIGINT, SIGTERM):
            with suppress(NotImplementedError):
                loop.add_signal_handler(sig, self._shutdown_event.set)

        if not self._http_session:
            if self._config.TLS_CA_PATH or (self._config.TLS_CERT_PATH and self._config.TLS_KEY_PATH):
                logger.info("Initializing SSL context for mTLS.")
                self._ssl_context = create_client_ssl_context(
                    ca_path=self._config.TLS_CA_PATH,
                    cert_path=self._config.TLS_CERT_PATH,
                    key_path=self._config.TLS_KEY_PATH,
                )
            connector = TCPConnector(ssl=self._ssl_context) if self._ssl_context else None
            self._http_session = ClientSession(connector=connector)
            if not self._clients:
                self._init_clients()

        # Connect transports
        await gather(*[client.connect() for _, client in self._clients])

        comm_task = create_task(self._manage_orchestrator_communications())

        token_rotation_task = None
        if self._ssl_context:
            token_rotation_task = create_task(self._manage_token_rotation())

        polling_task = create_task(self._start_polling())

        # Wait for shutdown signal
        await self._shutdown_event.wait()
        logger.info("Shutdown signal received. Starting graceful shutdown...")

        # 1. Stop polling and communications
        polling_task.cancel()

        # HLN RELIABILITY: Explicitly tell orchestrators we are going offline
        # This prevents new tasks from being dispatched to us during the drain period.
        logger.info("Sending final 'offline' heartbeat to all orchestrators...")
        with suppress(Exception):
            heartbeat = Heartbeat(
                worker_id=self._config.WORKER_ID,
                status="offline",
                load=float(self._current_load),
                current_tasks=list(self._active_tasks.keys()),
                supported_skills=[],
                hot_cache=list(self._hot_cache),
            )
            await gather(*[client.send_heartbeat(heartbeat) for _, client in self._clients], return_exceptions=True)

        comm_task.cancel()
        if token_rotation_task:
            token_rotation_task.cancel()

        # 2. Wait for active tasks with timeout (Drain Mode)
        if self._active_tasks:
            timeout = self._config.SHUTDOWN_TIMEOUT
            logger.info(f"Waiting for {len(self._active_tasks)} active tasks to complete (timeout: {timeout}s)...")
            try:
                await wait_for(
                    gather(*self._active_tasks.values(), return_exceptions=True), timeout=self._config.SHUTDOWN_TIMEOUT
                )
            except TimeoutError:
                logger.warning("Shutdown timeout reached. Some tasks may be interrupted.")
            except CancelledError:
                pass

        # 3. Final cleanup
        # Close transports
        await gather(*[client.close() for _, client in self._clients])

        if self._http_session and not self._http_session.closed and not self._session_is_managed_externally:
            await self._http_session.close()

    def run(self) -> None:
        """Runs the worker."""
        try:
            run(self.main())
        except KeyboardInterrupt:
            self._shutdown_event.set()

    async def _manage_token_rotation(self) -> None:
        """Periodically refreshes auth tokens for all clients."""
        await sleep(5)

        while not self._shutdown_event.is_set():
            min_expires_in = 3600

            for _, client in self._clients:
                try:
                    token_resp = await client.refresh_token()
                    if token_resp:
                        self._config.WORKER_TOKEN = token_resp.access_token
                        min_expires_in = min(min_expires_in, token_resp.expires_in)
                except Exception as e:
                    logger.error(f"Error in token rotation loop: {e}")

            refresh_delay = max(60, min_expires_in * 0.8)
            logger.debug(f"Next token refresh scheduled in {refresh_delay:.1f}s")
            await sleep(refresh_delay)

    async def _run_health_check_server(self) -> None:
        app = web.Application()

        async def health_handler(_):
            return web.Response(text="OK")

        app.router.add_get("/health", health_handler)
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, "0.0.0.0", self._config.WORKER_PORT)
        await site.start()
        await self._shutdown_event.wait()
        await runner.cleanup()

    def run_with_health_check(self) -> None:
        async def _main_wrapper() -> None:
            async with TaskGroup() as tg:
                tg.create_task(self._run_health_check_server())
                tg.create_task(self.main())

        try:
            run(_main_wrapper())
        except (KeyboardInterrupt, CancelledError):
            self._shutdown_event.set()
        except ExceptionGroup as eg:
            # Check if all exceptions in the group are related to shutdown
            if all(isinstance(e, (KeyboardInterrupt, CancelledError)) for e in eg.exceptions):
                self._shutdown_event.set()
            else:
                raise

    async def _start_websocket_manager(self) -> None:
        """Manages the command listeners."""
        listeners = []
        for _, client in self._clients:
            listeners.append(create_task(self._listen_to_single_transport(client)))

        await self._shutdown_event.wait()

        for listener in listeners:
            listener.cancel()

    async def _listen_to_single_transport(self, client: Transport) -> None:
        while not self._shutdown_event.is_set():
            try:
                async for command in client.listen_for_commands():
                    if command.command == COMMAND_CANCEL_TASK:
                        task_id = command.task_id
                        job_id = command.job_id
                        if task_id in self._active_tasks:
                            self._active_tasks[task_id].cancel()
                            logger.info(f"Cancelled task {task_id} (Job: {job_id or 'N/A'}) by orchestrator command.")
            except Exception as e:
                logger.error(f"Error in command listener: {e}")
            await sleep(5)



from typing import Any, Awaitable, Callable

from rxon.constants import (
    ERROR_CODE_DEPENDENCY as DEPENDENCY_ERROR,
)
from rxon.constants import (
    ERROR_CODE_INTERNAL as INTERNAL_ERROR,
)
from rxon.constants import (
    ERROR_CODE_INVALID_INPUT as INVALID_INPUT_ERROR,
)
from rxon.constants import (
    ERROR_CODE_PERMANENT as PERMANENT_ERROR,
)
from rxon.constants import (
    ERROR_CODE_RESOURCE_EXHAUSTED as RESOURCE_EXHAUSTED_ERROR,
)
from rxon.constants import (
    ERROR_CODE_SECURITY as SECURITY_ERROR,
)
from rxon.constants import (
    ERROR_CODE_TIMEOUT as TIMEOUT_ERROR,
)
from rxon.constants import (
    ERROR_CODE_TRANSIENT as TRANSIENT_ERROR,
)

Middleware = Callable[[dict[str, Any], Callable[[], Awaitable[Any]]], Awaitable[Any]]
CapacityChecker = Callable[[str], bool]


class ParamValidationError(Exception):
    pass


__all__ = [
    "INVALID_INPUT_ERROR",
    "PERMANENT_ERROR",
    "TRANSIENT_ERROR",
    "RESOURCE_EXHAUSTED_ERROR",
    "SECURITY_ERROR",
    "DEPENDENCY_ERROR",
    "TIMEOUT_ERROR",
    "INTERNAL_ERROR",
    "ParamValidationError",
]



import asyncio
from unittest.mock import MagicMock

import pytest
from rxon import Transport
from rxon.models import TaskPayload

from avtomatika_worker.worker import Worker


@pytest.mark.asyncio
async def test_validate_config_warns_on_unused_task_type_limits(mocker):
    """
    Tests that _validate_config warns when a task type limit is defined
    but no task with that type is registered.
    """
    logger_mock = mocker.patch("avtomatika_worker.worker.logger.warning")
    worker = Worker(task_type_limits={"gpu": 1})

    @worker.task("cpu_task")
    def cpu_task(params: dict):
        pass

    worker._validate_task_types()

    logger_mock.assert_called_once_with(
        "Configuration warning: A limit is defined for task type 'gpu', but no tasks are registered with this type."
    )


@pytest.mark.asyncio
async def test_debounced_heartbeat_sender(mocker):
    """
    Tests that _debounced_heartbeat_sender waits for the correct delay
    and then sends a heartbeat.
    """
    worker = Worker()
    worker._config.HEARTBEAT_DEBOUNCE_DELAY = 0.01
    mock_send_heartbeats = mocker.patch.object(worker, "_send_heartbeats_to_all", new_callable=mocker.AsyncMock)
    mock_sleep = mocker.patch("avtomatika_worker.worker.sleep", new_callable=mocker.AsyncMock)

    await worker._debounced_heartbeat_sender()

    mock_sleep.assert_called_once_with(0.01)
    mock_send_heartbeats.assert_called_once()


@pytest.mark.asyncio
async def test_schedule_heartbeat_debounce(mocker):
    """
    Tests that _schedule_heartbeat_debounce cancels a pending task and schedules a new one.
    """
    worker = Worker()
    mock_create_task = mocker.patch("avtomatika_worker.worker.create_task")

    # First call, no pending task
    worker._schedule_heartbeat_debounce()
    mock_create_task.assert_called_once()
    mock_create_task.call_args[0][0].close()  # Avoid RuntimeWarning

    # Second call, with a pending task
    pending_task = MagicMock()
    pending_task.done.return_value = False
    worker._debounce_task = pending_task
    worker._schedule_heartbeat_debounce()

    pending_task.cancel.assert_called_once()
    assert mock_create_task.call_count == 2
    mock_create_task.call_args[0][0].close()  # Avoid RuntimeWarning


@pytest.mark.asyncio
async def test_poll_for_tasks_receives_task(mocker):
    """Tests that _poll_for_tasks correctly processes a received task."""
    client = mocker.AsyncMock(spec=Transport)
    task_payload = TaskPayload(
        job_id="job-123", task_id="task-456", type="successful_task", params={"input": "test"}, tracing_context={}
    )
    client.poll_task.return_value = task_payload

    worker = Worker()

    # Register a dummy task so the worker is not "busy" (has supported tasks)
    @worker.task("successful_task")
    def dummy_handler(params):
        pass

    mocker.patch.object(worker, "_schedule_heartbeat_debounce")
    worker._process_task = mocker.AsyncMock()

    await worker._poll_for_tasks(client)

    client.poll_task.assert_called_once_with(timeout=worker._config.TASK_POLL_TIMEOUT)
    worker._process_task.assert_called_once()
    # verify client was NOT injected into task_data passed to _process_task as raw dict?
    # In new implementation, we convert TaskPayload to dict and inject client.
    call_args = worker._process_task.call_args[0][0]
    assert call_args["job_id"] == "job-123"
    assert call_args["client"] == client


@pytest.mark.asyncio
async def test_poll_for_tasks_no_task(mocker):
    """Tests that _poll_for_tasks handles no task correctly."""
    client = mocker.AsyncMock(spec=Transport)
    client.poll_task.return_value = None

    worker = Worker()

    # Register a dummy task so the worker is not "busy" (has supported tasks)
    @worker.task("dummy_task")
    def dummy_handler(params):
        pass

    mocker.patch.object(worker, "_schedule_heartbeat_debounce")
    worker._process_task = mocker.AsyncMock()

    await worker._poll_for_tasks(client)

    client.poll_task.assert_called_once()
    worker._process_task.assert_not_called()


@pytest.mark.asyncio
async def test_start_polling_busy(mocker):
    """Tests that _start_polling waits when the worker is busy."""
    worker = Worker()
    worker._get_current_state = mocker.MagicMock(return_value={"status": "busy"})
    mock_sleep = mocker.patch("avtomatika_worker.worker.sleep", new_callable=mocker.AsyncMock)
    mock_sleep.side_effect = asyncio.CancelledError  # To break the loop
    worker._registered_event.set()

    with pytest.raises(asyncio.CancelledError):
        await worker._start_polling()

    mock_sleep.assert_called_once_with(worker._config.IDLE_POLL_DELAY)


@pytest.mark.asyncio
async def test_start_polling_round_robin(mocker):
    """Tests that _start_polling uses round-robin to select orchestrators."""
    worker = Worker()
    worker._config.MULTI_ORCHESTRATOR_MODE = "ROUND_ROBIN"

    # Setup mock clients
    client1 = mocker.AsyncMock(spec=Transport)
    client2 = mocker.AsyncMock(spec=Transport)

    worker._clients = [
        ({"url": "http://test-1", "weight": 1, "current_weight": 0}, client1),
        ({"url": "http://test-2", "weight": 1, "current_weight": 0}, client2),
    ]
    worker._total_orchestrator_weight = 2
    worker._get_current_state = mocker.MagicMock(return_value={"status": "idle"})

    # We only want to test one iteration of the polling loop
    async def poll_side_effect(*args, **kwargs):
        worker._shutdown_event.set()

    worker._poll_for_tasks = mocker.AsyncMock(side_effect=poll_side_effect)
    worker._registered_event.set()

    await worker._start_polling()

    # With equal weights, the first one should be chosen
    worker._poll_for_tasks.assert_called_once_with(client1)


@pytest.mark.asyncio
async def test_process_task_exception(mocker):
    """Tests that _process_task handles exceptions in the task handler."""
    worker = Worker()

    client = mocker.AsyncMock(spec=Transport)

    @worker.task("failing_task")
    def failing_task(params: dict, **kwargs):
        raise ValueError("Task failed")

    task_data = {
        "job_id": "job-1",
        "task_id": "task-1",
        "type": "failing_task",
        "params": {},
        "tracing_context": {},
        "client": client,
    }

    await worker._process_task(task_data)

    client.send_result.assert_called_once()
    payload = client.send_result.call_args.args[0]
    assert payload.status == "failure"
    assert payload.error.code == "INTEGRITY_MISMATCH_ERROR"
    assert payload.error.message == "Task failed"


@pytest.mark.asyncio
async def test_process_unsupported_task(mocker):
    """Tests that _process_task handles unsupported tasks correctly."""
    worker = Worker()
    client = mocker.AsyncMock(spec=Transport)

    task_data = {
        "job_id": "job-1",
        "task_id": "task-1",
        "type": "unsupported_task",
        "params": {},
        "tracing_context": {},
        "client": client,
    }

    await worker._process_task(task_data)

    client.send_result.assert_called_once()
    payload = client.send_result.call_args.args[0]
    assert payload.status == "failure"
    assert payload.error.message == "Unsupported task: unsupported_task"


@pytest.mark.asyncio
async def test_process_task_cancelled(mocker):
    """Tests that _process_task handles cancelled tasks correctly."""
    worker = Worker()
    client = mocker.AsyncMock(spec=Transport)

    @worker.task("cancellable_task")
    async def cancellable_task(params: dict, **kwargs):
        raise asyncio.CancelledError

    task_data = {
        "job_id": "job-1",
        "task_id": "task-1",
        "type": "cancellable_task",
        "params": {},
        "tracing_context": {},
        "client": client,
    }

    with pytest.raises(asyncio.CancelledError):
        await worker._process_task(task_data)

    client.send_result.assert_called_once()
    payload = client.send_result.call_args.args[0]
    assert payload.status == "cancelled"

    @pytest.mark.asyncio
    async def test_run_with_health_check(mocker):
        """
        Tests that run_with_health_check runs the health check server and the main worker loop.
        """
        worker = Worker()
        mock_run = mocker.patch("avtomatika_worker.worker.run")
        worker._run_health_check_server = mocker.AsyncMock()
        worker.main = mocker.AsyncMock()

        # Create a separate async function to call the method
        async def run_test():
            worker.run_with_health_check()

        # Run the test function
        await run_test()

        # Check that asyncio.run was called
        mock_run.assert_called_once()

        # To check the inner calls, we need to get the coroutine passed to asyncio.run
        wrapper_coro = mock_run.call_args[0][0]
        await wrapper_coro

        # Verify that both server and main were called
        worker._run_health_check_server.assert_called_once()
        worker.main.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:coroutine 'AsyncMockMixin._execute_mock_call' was never awaited")
async def test_run_health_check_server(mocker):
    """
    Tests that _run_health_check_server starts and stops the aiohttp server.
    """
    worker = Worker()

    # Explicitly mock classes and their return values for better control
    mock_app_instance = mocker.MagicMock()
    mock_app_class = mocker.patch("aiohttp.web.Application", return_value=mock_app_instance)

    mock_runner_instance = mocker.MagicMock()
    mock_runner_instance.setup = mocker.AsyncMock()
    mock_runner_instance.cleanup = mocker.AsyncMock()
    mock_runner_class = mocker.patch("aiohttp.web.AppRunner", return_value=mock_runner_instance)

    mock_site_instance = mocker.MagicMock()
    mock_site_instance.start = mocker.AsyncMock()
    mock_site_class = mocker.patch("aiohttp.web.TCPSite", return_value=mock_site_instance)

    # Set the shutdown event to stop the server
    async def wait_side_effect(*args, **kwargs):
        worker._shutdown_event.set()

    worker._shutdown_event.wait = mocker.AsyncMock(side_effect=wait_side_effect)

    await worker._run_health_check_server()

    mock_app_class.assert_called_once_with()  # Application()
    mock_app_instance.router.add_get.assert_called_once_with(
        "/health", mocker.ANY
    )  # Assuming internal call to app.router.add_get

    mock_runner_class.assert_called_once_with(mock_app_instance)  # AppRunner(app)
    mock_runner_instance.setup.assert_called_once()
    mock_runner_instance.cleanup.assert_called_once()

    mock_site_class.assert_called_once_with(
        mock_runner_instance, "0.0.0.0", worker._config.WORKER_PORT
    )  # TCPSite(runner, ...)
    mock_site_instance.start.assert_called_once()



import sys
from unittest.mock import patch

import pytest

from avtomatika_worker.config import WorkerConfig


@pytest.mark.asyncio
async def test_s3_missing_dependency():
    """
    Tests that S3Manager raises appropriate errors when 'obstore' is not installed.
    """
    # Remove the module from sys.modules to force reload
    if "avtomatika_worker.s3" in sys.modules:
        del sys.modules["avtomatika_worker.s3"]

    # Simulate import error for obstore
    with patch.dict(sys.modules, {"obstore": None}):
        # Re-import the module under test
        import avtomatika_worker.s3 as s3_module

        assert s3_module._HAS_S3 is False

        config = WorkerConfig()
        # Ensure S3 is "enabled" in config so logic tries to proceed
        config.S3_ENDPOINT_URL = "http://localhost:9000"

        manager = s3_module.S3Manager(config)

        # 1. Check explicit check
        with pytest.raises(RuntimeError) as excinfo:
            manager._check_availability()
        assert "install 'avtomatika-worker[s3]'" in str(excinfo.value)

        # 2. Check process_params (async)
        with pytest.raises(RuntimeError):
            await manager.process_params({"file": "s3://bucket/file"}, "task-1")

    # Cleanup: remove the modified module so subsequent tests import the correct one
    if "avtomatika_worker.s3" in sys.modules:
        del sys.modules["avtomatika_worker.s3"]
